use actix_web::{
    get, post,
    web::{self, BufMut},
    App, HttpRequest, HttpResponse, HttpServer, Responder, HttpResponseBuilder, http::StatusCode,
};
use arti_client::{TorClient, TorClientConfig};
use eyre::{ErrReport, Result};
use tokio::io::{AsyncReadExt, AsyncWriteExt};

mod types;

struct State {
    tor_client: TorClient<tor_rtcompat::PreferredRuntime>,
}

#[get("/{tail:.*}")]
async fn get(req: HttpRequest, data: web::Data<State>, path: web::Path<String>) -> impl Responder {
    let tail = "/".to_string() + &path.into_inner();

    let host = match req.uri().host() {
        Some(v) => v.replace(".vanten-s.com", ""),
        None => return HttpResponse::InternalServerError().body("How did you manage to get an invalid URI??")
    };

    let raw_website_response = match get_website(&host, 80, &req, None, &data.tor_client).await {
        Ok(v) => v,
        Err(e) => return HttpResponse::InternalServerError().body(format!("hihi, {e}")),
    };

    HttpResponse::Ok().body("Hi!")
}

#[tokio::main]
async fn main() -> Result<()> {
    let config = TorClientConfig::default();

    let tor_client = TorClient::create_bootstrapped(config).await?;
    let server = HttpServer::new(move || App::new().app_data(State { tor_client: tor_client.clone() }).service(get))
        .bind(("127.0.0.1", 8080))?
        .run();
    let server = server.await;

    println!("Everything Started!");

    Ok(())
}

async fn get_website(
    address: &str,
    port: u16,
    request: &HttpRequest,
    body: Option<String>,
    tor_client: &TorClient<tor_rtcompat::PreferredRuntime>,
) -> Result<Vec<u8>> {
    let mut stream = tor_client.connect((address, port)).await?;
    stream.write_all(&request_to_bytes(request, body)?).await?;
    stream.flush().await?;

    println!("Sent request to {address}");
    let mut response_status: Vec<u8> = Vec::new();

    loop {
        let response = stream.read_u8().await?;
        response_status.push(dbg!(response));
        if response == b'\n' {
            break;
        }
    }

    let _ = dbg!(String::from_utf8(response_status));

    todo!();
}

fn request_to_bytes(request: &HttpRequest, body: Option<String>) -> Result<Vec<u8>> {
    let mut finished = vec![];

    // First Line
    {
        use actix_web::http::Version as V;
        let method = request.method().to_string().as_bytes();
        let path = request.path().as_bytes();
        let version = match request.version() {
            V::HTTP_09 => "HTTP/0.9",
            V::HTTP_10 => "HTTP/1",
            V::HTTP_11 => "HTTP/1.1",
            V::HTTP_2 => "HTTP/2",
            V::HTTP_3 => "HTTP/3",
        }
        .as_bytes();

        finished.put_slice(method);
        finished.push(b' ');
        finished.put_slice(path);
        finished.push(b' ');
        finished.put_slice(version);
        finished.push(b'\n');
    }

    // Headers
    {
        for header in request.headers() {
            let name = header.0.to_string().as_bytes();
            let value = header.1.as_bytes();
            finished.put_slice(name);
            finished.put_slice(b": ");
            finished.put_slice(value);
            finished.push(b'\n');
        }
        finished.push(b'\n');
    }

    match body {
        Some(v) => finished.put_slice(v.as_bytes()),
        None => {},
    }

    Ok(finished)
}
